
#include <stdlib.h>
#include <assert.h>

#include <iostream>
#include <limits>
#include <algorithm>

#include <LPBoost.h>

namespace LPBoost
{

void
LPBoost::BuildClassifier (std::vector<const SeqMine::Sequence*>& S,
	std::vector<double>& Ylabels)
{
	/* Initialize Wcur uniformly, but class-balanced, such that the weights
	 * sum to one.
	 */
	std::vector<double> Wcur (Ylabels.size ());
	unsigned int posCount = std::count (Ylabels.begin (), Ylabels.end (), 1.0);
	unsigned int negCount = std::count (Ylabels.begin (), Ylabels.end (), -1.0);
	for (unsigned int n = 0 ; n < Wcur.size () ; ++n)
		Wcur[n] = 1.0 / (2.0 * ((Ylabels[n] >= 0.0) ? ((double) posCount)
			: ((double) negCount)));

	/* nu regularization parameter
	 */
	double D = 1.0 / (nu * Ylabels.size ());

	/* Variables: [lambda, gamma]
	 */
	lprec* lp = make_lp (0, Ylabels.size () + 1);
	assert (lp != NULL);
	set_lp_name (lp, "LPBoost 2-class dual problem");

	/* LP_SOLVE 5.5.10 has problems with some models we generate.  Using the
	 * primal-primal variant for both phases of the solution (feasibility and
	 * optimization).
	 *
	 * The default strategy would be SIMPLEX_DUAL_PRIMAL
	 */
	set_simplextype (lp, SIMPLEX_PRIMAL_PRIMAL);

	/* Add the \sum_{n=1}^{\ell} \lambda_n = 1 constraint
	 */
	REAL lambdasum[1 + Ylabels.size () + 1];
	lambdasum[0] = lambdasum[1+Ylabels.size()] = 0.0;
	for (unsigned int n = 0 ; n < Ylabels.size () ; ++n)
		lambdasum[1+n] = 1.0;
	add_constraint (lp, lambdasum, GE, 1.0);

	/* Number of stumps in the LP so far
	 */
	unsigned int stump_count = 0;
	stumps.clear ();

	/* Set objective function: f(lambda,gamma) = gamma
	 *
	 * the variable ordering we use is [lambda, gamma]
	 *
	 * FIXME: check if this is allowed (first setting the objective
	 * function, then adding constraints)  It says in the documentation
	 * this way is more efficient
	 */
	REAL obj[1 + Ylabels.size () + 1];
	for (unsigned int n = 0 ; n < (1+Ylabels.size ()) ; ++n)
		obj[n] = 0.0;	// lambda_n

	obj[1+Ylabels.size()] = 1.0;	// gamma
	set_obj_fn (lp, obj);
	set_minim (lp);	// Set objective as minimization problem

	/* Set rather expensive presolve operations in order to cope with an
	 * otherwise possibly numerically degenerate LP.
	 */
#if 0
	// PROBLEM: presolve also changes our model considerably, hence using the
	// model itself as a storage for our data does not work.
	// Alternative: rebuild the model every iteration or use copy_lp.
	set_presolve (lp, PRESOLVE_ROWS | PRESOLVE_COLS | PRESOLVE_LINDEP
		| PRESOLVE_IMPLIEDFREE | PRESOLVE_ROWDOMINATE
		| PRESOLVE_COLDOMINATE | PRESOLVE_MERGEROWS
		| PRESOLVE_DUALS, get_presolveloops (lp));
#endif

	/* theta is the minimum required gain (see lpboost.m)
	 */
	double theta = std::numeric_limits<double>::min ();
	//double stabilizationWindowSize = 0.125 * (1.0 / ((double) Ylabels.size ()));
	double stabilizationWindowSize = 0.125 * (1.0 / ((double) Ylabels.size ()));

	for (unsigned int iter = 0 ; true ; ++iter) {
		std::cout << "LPBoost 2-class, iter " << iter
			<< ", theta " << theta << std::endl;

		SeqMine::PrefixSpanLPBoost boost (Ylabels, Wcur, theta,
			pricingCount, 100000, minsup, minlen, maxlen);
		boost.Mine (S);

		/* In case no patterns were found, we are done.
		 */
		if (boost.bestPatterns.size () == 0)
			break;

		/* Mining returned with some successful sequences that gain us
		 * something.  Lets take a look at those and where they appear.
		 */
		std::cout << "Obtained sequences" << std::endl;
		for (std::multimap<double, std::pair<SeqMine::SequenceT, double> >::iterator
			iv = boost.bestPatterns.begin () ; iv != boost.bestPatterns.end () ; ++iv)
		{
			std::cout << "gain " << iv->first << " (ybase "
				<< iv->second.second << "), seq: "
				<< iv->second.first << std::endl;

			std::cout << "   occurence: " << std::endl;
			for (unsigned int n = 0 ; n < Ylabels.size () ; ++n) {
				bool sampleActive = (iv->second.first.Occurence.find (n)
					!= iv->second.first.Occurence.end ());

				if (sampleActive == false)
					continue;

				std::cout << n << "(" << Ylabels[n] << ") ";
			}
			std::cout << std::endl;
		}

		/* Convergence check: the new best sequence has to be better than our
		 * current performance.
		 */
		std::cout << "best gain: " << boost.bestPatterns.begin ()->first << std::endl;
		if (boost.bestPatterns.begin ()->first <= theta + convergenceEpsilon) {
			std::cout << "Converged, there is no better pattern." << std::endl;

			break;
		}

		/* Add the new stumps to the stump set H
		 */
		stump_count += boost.bestPatterns.size ();
		resize_lp (lp, 1 + stump_count, Ylabels.size () + 1);

		/* Add box constraints:
		 *    0 <= \lambda_n <= D, for all n,
		 *
		 * Also add a stabilizing trust region constraint in the second and
		 * following iterations.
		 */
		for (unsigned int n = 0 ; n < Ylabels.size () ; ++n) {
			double lambdaLow = 0.0;
			double lambdaHigh = D;

			double stabLow = Wcur[n] - stabilizationWindowSize;
			double stabHigh = Wcur[n] + stabilizationWindowSize;
			if (stabLow > 0.0)
				lambdaLow = stabLow;

			if (stabHigh < D)
				lambdaHigh = stabHigh;

			set_bounds (lp, 1+n, lambdaLow, lambdaHigh);	// 0 <= lambda <= D
		}

		/* Explicitly we relax gamma to be between -Inf and Inf, because the
		 * default would be gamma >= 0.
		 */
		REAL inf = get_infinite (lp);
		set_bounds (lp, 1+Ylabels.size (), -inf, inf);	// gamma
		//set_bounds (lp, 1+Ylabels.size (), 0, inf);	// gamma

		/* Add the mined patterns (new constraints)
		 */
		set_add_rowmode (lp, TRUE);	// Turn row mode on (faster adding of constraints)
		for (std::multimap<double, std::pair<SeqMine::SequenceT, double> >::iterator
			iv = boost.bestPatterns.begin () ; iv != boost.bestPatterns.end () ; ++iv)
		{
			/* Add to global list of stumps
			 */
			stumps.push_back (iv->second);

			REAL row[1+Ylabels.size()+1];
			row[0] = 0.0;

			for (unsigned int n = 0 ; n < Ylabels.size () ; ++n) {
				/* Predicate: does sample n contain the current pattern?
				 */
				bool sampleActive = (iv->second.first.Occurence.find (n)
					!= iv->second.first.Occurence.end ());
				double stumpResponse =
					(sampleActive ? 1.0 : -1.0) * iv->second.second;

				row[1+n] = Ylabels[n] * stumpResponse;
			}
			row[1+Ylabels.size()] = -1.0;	// -gamma
			add_constraint (lp, row, LE, 0.0);
		}
		set_add_rowmode (lp, FALSE);

		/* Dump the whole LP to stdout
		 */
		print_lp (lp);

		char lpfilename[256];
		sprintf (lpfilename, "lp-debug-%d.lp", iter);
		write_lp (lp, lpfilename);

		//set_verbose (lp, CRITICAL);
		set_verbose (lp, NORMAL);
		int res = solve (lp);
		assert (res == 0);

		print_duals (lp);
		print_solution (lp, Ylabels.size()+1);

		/* The solution contains the primal variables (lambda, gamma), as
		 * well as the dual variables (the original primal variables, but we
		 * are only interested in alpha and rho, where we get alpha directly
		 * from the Lagrangian multipliers of the constraints and rho from the
		 * objective function)
		 */
		//int nrows = get_Nrows (lp);
		int nrows = 1+stump_count;
		REAL varRes[1+nrows+Ylabels.size()+1];	// 1+constraints + variables
		//get_primal_solution (lp, varRes);
		for (unsigned int n = 0 ; n < 1+nrows+Ylabels.size()+1 ; ++n)
			varRes[n] = get_var_primalresult (lp, n);

#if 0
		for (unsigned int n = 0 ; n < 1+nrows+Ylabels.size()+1 ; ++n)
			std::cout << "varRes[" << n << "] = " << varRes[n] << std::endl;
#endif

		/* Alpha, the stump multipliers
		 */
		alpha.resize (stump_count);
		//REAL* ptr_duals;
		//get_ptr_dual_solution (lp, &ptr_duals);
		std::cout << "alpha:";
		for (unsigned int n = 0 ; n < stump_count ; ++n) {
			//alpha[n] = -ptr_duals[1+1+n];
			alpha[n] = -get_var_dualresult (lp, 1+1+n);
			std::cout << " " << alpha[n];
		}
		std::cout << std::endl;

#if 0
		for (unsigned int n = 0 ; n < stump_count ; ++n) {
			alpha[n] = -varRes[1+1+n];
			std::cout << "alpha_" << n << " = " << alpha[n] << std::endl;
		}
#endif

		std::cout << "weights:";
		for (unsigned int n = 0 ; n < Ylabels.size () ; ++n) {
			Wcur[n] = varRes[1+nrows+n];
			std::cout << " " << Wcur[n] << "(" << Ylabels[n] << ")";
		}
		std::cout << std::endl;

		theta = get_var_primalresult (lp, 1+nrows+Ylabels.size());
		std::cout << "theta = " << theta << std::endl;
		//assert (0);

		PrintCurrentTrainingStatus (Ylabels);
	}
	delete_lp (lp);
}


double
LPBoost::ValidationLoss (void) const
{
	assert (doValidation);

	double loss = 0.0;
	std::vector<double> sOut (Yval.size ());

	/* Sum the classifier output.
	 */
	for (unsigned int sn = 0 ; sn < alpha.size () ; ++sn) {
		SeqMine::SequenceTSearch st (stumps[sn].first);

		for (unsigned int n = 0 ; n < Yval.size () ; ++n)
			sOut[n] += alpha[sn] * stumps[sn].second *
				(st.containsSequence (*Sval[n]) ? 1.0 : -1.0);
	}

	/* Compute the empirical loss
	 */
	for (unsigned int n = 0 ; n < Yval.size () ; ++n) {
		double margin = sOut[n] * Yval[n];
		if (margin <= 1e-4)
			loss += 1.0;
	}
	loss /= (double) Yval.size ();

	return (loss);
}

void
LPBoost::SetValidationSet (std::vector<const SeqMine::Sequence*>& Sval,
	std::vector<double>& Yval)
{
	doValidation = true;

	this->Sval = Sval;
	this->Yval = Yval;
}

void
LPBoost::PrintCurrentTrainingStatus (std::vector<double>& Ylabels) const
{
	double trainingLoss = 0.0;

	for (unsigned int n = 0 ; n < Ylabels.size () ; ++n) {
		std::cout << "sample " << n << ": ";

		double sum = 0.0;
		for (unsigned int d = 0 ; d < alpha.size () ; ++d) {
			std::cout << " " << alpha[d] << "*";

			bool active = (stumps[d].first.Occurence.find (n) !=
				stumps[d].first.Occurence.end ());
			double response = stumps[d].second * (active ? 1.0 : -1.0);
			std::cout << response;

			sum += alpha[d] * response;
		}
		std::cout << " = " << sum << " (true " << Ylabels[n] << ")";

		if (sum*Ylabels[n] <= 1e-4) {
			std::cout << " INCORRECT";
			trainingLoss += 1.0;
		}
		std::cout << std::endl;
	}
	trainingLoss /= Ylabels.size ();
	std::cout << "TOTAL TRAINING LOSS: " << trainingLoss << std::endl;

	if (doValidation)
		std::cout << "TOTAL VALIDATN LOSS: " << ValidationLoss () << std::endl;
}

}


